"lm.dibp" <-
function( x, y, formula1=y1y2~., formula2=y3~1, data, zeroL3=FALSE, distribution='discrete', jmax=2, maxit=300,  pres=1e-8, print.details=FALSE)
{
options(warn=-1)
#
# ------------------------------------------------------------------------------------
# Karlis and Ntzoufras (2003, 2004)
# EM algorithms for Bivariate Poisson Models
# ------------------------------------------------------------------------------------
#
# PARAMETERS COMMON WITH lm.bp
# x        : name of first response variable (character object)
# y        : name of second response variable (character object)
# formula1 : formula for the modeling of lambda1 and lambda2
# formula2 : formula for the modeling of lambda3 
# data     : data.frame which contains data {required arguement}
# zeroL3   : Logical argument controlling whether lambda3 is zero (DblPoisson) or not
# maxit    : maximum number of iterations 
# pres     : precision of the relative likelihood difference after which EM stops
# print.details: Logical argument controlling whether beta parameters will we 
#                printed while EM runs. 
#
# PARAMETERS ADDITIONAL TO lm.bp
# distribution : Selection of diagonal inflation distribution. 
#                Three choices are available: 
#                ='discrete'  : Discrete, jmax is the number of diagonal elements [0,1,...,]
#                ='poisson'   : Poisson with mean theta.  
#                ='geometrics': Geometric with success probability theta. 
#                Default is DISCRETE(2). theta[1] and theta[2] stand for theta_1, theta_2
#                                        while theta_0=1-theta[1]-theta[2].
# jmax         : Used only for DISCRETE diagonal distribution (distribution='discrete').
#                Indicates the number of parameters of the DISCRETE distribution.
# ------------------------------------------------------------------------------------
#
#
#
# set full model
#
newformula<-sub( '\\.','noncommon:(.-l1-l2)', as.character(formula1[3])  )
formula1<-formula( paste('y1y2~',newformula,sep='') )
#
# error messages if x,y are not labels
if( !(is.character(x)) ){ print('ERROR: x should be the name of the 1st response included in quotes.') }
if( !(is.character(y)) ){ print('ERROR: y should be the name of the 2nd response included in quotes.') } 
namex<-x
namey<-y
x<-data[,names(data)==namex]
y<-data[,names(data)==namey]
#
#
# Data length
n<-length(x)
lengthprintvec<-1
#
#
#
# definition of diagonal inflated distribution
	maxy<-max(c(x,y))
#
#  changing distribution to codes 1,2,3	
	dist<-distribution
	if      ( charmatch( dist, 'poisson'  , nomatch=0) ==1 ) {distribution<-2} 
	else if ( charmatch( dist, 'geometric', nomatch=0) ==1 ) {distribution<-3}	
	else if ( charmatch( dist, 'discrete' , nomatch=0) ==1 ) {distribution<-1}
	if ( distribution==1 ){ 
			dilabel<-paste('Inflation Distribution: Discrete with J=',jmax)
			if (jmax==0) {theta<-0}
			else { theta<-1:jmax*0+1/(jmax+1) }
			di.f<-function (x, theta){
							JMAX<-length(theta)
							if      (x>JMAX) { res<-0 }
							else if (x==0)   { res<-1-sum(theta) }
							else             { res<-theta[x] } 
							res
					 }
	}
	else if ( distribution==2 ){ 
			dilabel<-'Inflation Distribution: Poisson'
			theta<-1.0;
			di.f<-function (x, theta){
							if (theta>0) { res<-dpois( x, theta ) }
							else {
								if (x==0) { res<-1}
								else {res<-1e-12}
							}
					 }
	}								 
	else if ( distribution==3 ){ 
			dilabel<-'Inflation Distribution: Geometric'
			theta<-0.5;
			di.f<-function (x, theta){
							if (theta>0) { 
								if(theta==1) {theta<-0.9999999}
								res<-dgeom( x, theta ) }
							else if (theta==1){
								if (x==0) { res<-1}
								else {res<-1e-12}
							}
							else {res<-1e-12}
					 }
	}
	else {
		print( paste(distribution, 'Not known distribution.', sep=': ') )
	}
# ------
# setting up data frames, vectors and data
#
# form dataframes used
# data1 includes modelling on lambda1 and lambda2
# data2 includes modelling on lambda3 
#
data1<-data.frame( y1y2=c( x, y) )
data2<-data.frame( y3 = rep(0, n ) )
#
p<-length(as.data.frame(data))
data1[,2:(p+1)]<-rbind(data, data) 
names(data1)[2:(p+1)]<-names(data)
#
# removing x and y
data1<-data1[ , names(data1)!=namex]
data1<-data1[ , names(data1)!=namey]
#
# noncommon vector
data1$noncommon<- as.factor(c(1:n*0,1:n*0+1))
contrasts(data1$noncommon)<-contr.treatment(2, base=1)
#
# vectors l1 and l2
data1$l1<-c(1:n*0+1,1:n*0  )
data1$l2<-c(1:n*0  ,1:n*0+1)
if (!zeroL3){	
		data2[,2:(p-1)]<-data1[1:n,2:(p-1)]
		names(data2)[2:(p-1)]<-names(data1)[2:(p-1)]
		}
#####
tmpform1<-as.character(formula1[3])
while( regexpr('c\\(',tmpform1) != -1)
{	
	temppos1<-regexpr('c\\(',tmpform1)[1]
	tempfor <-substring( tmpform1, first = temppos1+ 2 )
	temppos2<-regexpr('\\)' , tempfor)[1]
	tempvar <-substring( tempfor , first = 1,  last = temppos2-1 )
	temppos3<-regexpr(', ' , tempvar)[1]
	tempname1<-substring(tempfor    , first = 1,  last = temppos3-1 )
	tempname2<-substring(tempfor    , first = temppos3+2, last=temppos2-1)
	tempname2<-sub( '\\)','', tempname2  )
	tempvar1<-data[, names(data)==tempname1]
	tempvar2<-data[, names(data)==tempname2]
	data1$newvar1<-c(tempvar1, tempvar2)
#
	if( is.factor(tempvar1)& is.factor(tempvar2) ){
		data1$newvar1<-as.factor(data1$newvar1)	
		if (all(levels(tempvar1)==levels(tempvar2))){
			attributes(data1$newvar1)<-attributes(tempvar1)}
	}			
	tempvar<-sub( ', ' , '..',  tempvar )
	names(data1)[names(data1)=='newvar1']<-tempvar
	newformula<-sub( 'c\\(','', tmpform1  )
	newformula<-sub( '\\)','', newformula  )
	newformula<-sub( ', ' , '..', newformula  )
	tmpform1<-newformula
}
formula1<-formula( paste('y1y2~',newformula, sep='' ) )
#####
rm(temppos1)
rm(temppos2)
rm(temppos3)
rm(tmpform1)
rm(tempfor)
rm(tempvar)
rm(tempvar1)
rm(tempvar2)
rm(tempname1)
rm(tempname2)
# ------
# initial values for parameters
prob<-0.20
s<-rep(0,n)
vi<-1:n*0
v1<-1-c(vi,vi)
like<-1:n*0
zero<- ( x==0 )|( y==0 )
if   (zeroL3) { lambda3<-rep(0,n) } 
else          { lambda3<-rep( max(0.1, cov(x,y,use='complete.obs')), n) } 
#
#
#
#
# Initial values for lambda

data1$v1<-1-c(vi,vi);

lambda<-glm( formula1, family=poisson, data=data1, weights=v1, maxit=100)$fitted
#
lambda1<-lambda[1:n]
lambda2<-lambda[(n+1):(2*n)]
#
difllike<-100.0
loglike0<-1000.0
i<-0
ii<-0

if (zeroL3) {
	#
	# fit double poisson diagonal inflated model
	loglike<-rep(0, maxit)
	lambda3<-1:n*0
	while ( (difllike>pres) && (i <= maxit) ) {
	i<-i+1
	#####   E step  ######
	for (j in 1:n) {
		if (zero[j]) {
			s[j]<-0;		
#			calculation of log-likelihood			
			if (x[j]==y[j]) {
				density.di<-di.f( 0.0, theta )
				like[j]<-log( (1-prob)*exp(-lambda1[j]-lambda2[j])+prob*density.di );
				vi[j]<-prob*density.di*exp(-like[j]) }
			else{
			        like[j]<-log(1-prob)+log(dpois(x[j],lambda1[j])) 																					 +log(dpois(y[j],lambda2[j]));
				vi[j]<-0.0 ;}
		}			
		else {
			if (x[j]==y[j]) {
					density.di<-di.f( x[j],theta );
					like[j]<-log( (1-prob)*dpois( x[j],lambda1[j] )*dpois( y[j],lambda2[j] ) + prob*density.di );
					vi[j]  <- prob*density.di*exp( -like[j] ) }
			else	{
					vi[j]<-0.0;
					like[j]<-log(1-prob)+log( dpois(x[j],lambda1[j])*dpois(y[j],lambda2[j]) )}
		}
	}
#### end of E-step #########
	x1<-x;
	x2<-y;
	loglike[i]<-sum( like ) ; 
	difllike<-abs( (loglike0-loglike[i])/loglike0 )
	loglike0<-loglike[i]
	#
	#
########### M-step ############
	# estimate mixing proportion
	prob<-sum(vi)/n
	#
	# maximization of each theta parameter
	if ( distribution == 1 ) {
		# calculation of theta_j, j=1,...,jmax ; theta_0=1-sum(theta)
		if (jmax==0) { theta<-0 }
		else {
			for (ii in 1:jmax) {
				temp<-as.numeric(( (x==ii) & (y==ii) ));
				theta[ii]<-sum(temp*vi)/sum(vi) 
				}
			}
	}			
	else if (distribution==2){		
			# calculation of theta for poisson diagonal inflation
			theta<- sum(vi*x)/sum(vi) }
	else if (distribution==3){		
			# calculation of theta for geometric diagonal inflation
			theta<- sum(vi)/( sum(vi*x)+sum(vi) ) }
	#
	#	fit model on lambda1 & lambda2
	#
	data1$v1<- 1-c(vi,vi);
	data1$v1[ (data1$v1<0)&(data1$v1>-1.0e-10) ]<-0.0
#	
	x1[(x1<0)&(x1>-1.0e-10)]<-0.0
	x2[(x2<0)&(x2>-1.0e-10)]<-0.0
	data1$y1y2<-c(x1,x2)
	m<-glm( formula1, family=poisson, data=data1, weights=v1 , maxit=100)

	p3<-length(m$coef)
	beta<-m$coef 
# ----------------------------------------------------
#	creating names for parameters
	names(beta)<-newnamesbeta( beta )
#
# 	end of name creations (l1, l2, l2-l1, blank)
#  -------------------------------------------------
	betaparameters<-splitbeta( beta )
#	
	lambda<-fitted(m)
	lambda1<-lambda[1:n]
	lambda2<-lambda[(n+1):(2*n)]
	#
	#####   end of M step  ######
#
#  printing also beta
	if (print.details) {
		printvec<- c( i,beta,100.0*prob, theta,  loglike[i], difllike );	
		names(printvec)<-c( 'iter', names(beta),  'Mix.p(%)', paste( 'theta', 1:length(theta),sep='' ), 'loglike', 'Rel.Dif.loglike')
		}
#  limited print out
	else {		
		printvec<- c( i, 100.0*prob, theta,  loglike[i], difllike );	
		names(printvec)<-c( 'iter','Mix.p(%)', paste( 'theta', 1:length(theta),sep='' ), 'loglike', 'Rel.Dif.loglike')
	}
	lengthprintvec<-length(printvec)
	print.default( printvec, digits=4 )
	}

#
#	calculation of BIC and AIC for double poisson model
	if ( (distribution==1)&&(jmax==0) ){noparams<- m$rank +1}
	else                              {noparams<- m$rank + length( theta ) +1}
	AIC<- -2*loglike[i] + noparams * 2
	BIC<- -2*loglike[i] + noparams * log(2*n)
#	
#		
#	Calculation of BIC, AIC of Poisson saturated model
	x.mean<-x
	x.mean[x==0]<-1e-12
	y.mean<-y
	y.mean[y==0]<-1e-12
	AIC.sat <-  sum(log( dpois( x , x.mean ) ) + log( dpois( y , y.mean ) ))
	BIC.sat <-  -2 * AIC.sat + (2*n)* log(2*n)
	AIC.sat <-  -2 * AIC.sat + (2*n)* 2
#
#
	AICtotal<-c(AIC.sat, AIC); 
	BICtotal<-c(BIC.sat, BIC );
	names(AICtotal)<-c('Saturated', 'DblPois')
	names(BICtotal)<-c('Saturated', 'DblPois')
#
	allbeta<-c(betaparameters$beta1,betaparameters$beta2)
	names(allbeta)<-c( paste( '(l1):', names(betaparameters$beta1), sep='' ),paste('(l2):', names(betaparameters$beta2), sep='' ) )
#
# 	calculation of fitted values
#  ----------------------------
	fittedval1<-(1-prob)*m$fitted[1:n]
	fittedval2<-(1-prob)*m$fitted[(n+1):(2*n)]
#  	
	meandiag<-0
	if ((distribution==1)&&(jmax>0)) { meandiag<-sum( theta[1:jmax]*1:jmax ) }
	else if (distribution==2) { meandiag<-theta }
	else if (distribution==3) { meandiag<- (1-theta)/theta }
#	
	fittedval1[x==y]<-prob*meandiag + fittedval1[x==y]
	fittedval2[x==y]<-prob*meandiag + fittedval2[x==y]
#
	result<-list(beta=allbeta, beta1=betaparameters$beta1, beta2=betaparameters$beta2, 			
				p=prob, theta=theta, diagonal.distribution=dilabel,
				lambda1=m$fitted[1:n], lambda2=m$fitted[(n+1):(2*n)],
				fitted1=fittedval1, fitted2=fittedval2, loglikelihood=loglike[1:i], 
				parameters=noparams, AIC=AICtotal, 	BIC=BICtotal,iterations=i )
#
#
# end of diagonal inflated double poisson model
}
else {	
	loglike<-rep(0,maxit)
	while ( (difllike>pres) && (i <= maxit) ) {
	i<-i+1
	#####   E step  ######
	for (j in 1:n) {
		if (zero[j]) {
			s[j]<-0;		
#			calculation of log-likelihood			
			if (x[j]==y[j]) {
				density.di<-di.f( 0.0, theta )
				like[j]<- log( (1-prob)*exp(-lambda1[j]-lambda2[j]-lambda3[j])+prob*density.di );
				vi[j]<-prob*density.di*exp(-like[j]) }
			else{
			        like[j]<-log(1-prob)-lambda3[j] +log(dpois(x[j],lambda1[j])) 									+log(dpois(y[j],lambda2[j]));
				vi[j]<-0.0 ;}
		}			
		else {
			lbp1<-pbivpois(x[j]-1, y[j]-1, lambda=c(lambda1[j],lambda2[j],lambda3[j]),  log=TRUE );
			lbp2<-pbivpois(x[j]  , y[j]  , lambda=c(lambda1[j],lambda2[j],lambda3[j]), log=TRUE );	
			s[j]<-exp( log(lambda3[j]) + lbp1 - lbp2 );
#			like[j]<-lbp2;
			if (x[j]==y[j]) {
					density.di<-di.f( x[j],theta );
					like[j]<-log( (1-prob)*exp(lbp2) + prob*density.di );
					vi[j]  <- prob*density.di*exp( -like[j] ) }
			else	{
					vi[j]<-0.0;
					like[j]<-log(1-prob)+lbp2 }
		}
	}
#### end of E-step #########
	x1<-x-s;
	x2<-y-s;
	loglike[i]<-sum( like ) ; 
	difllike<-abs( (loglike0-loglike[i])/loglike0 )
	loglike0<-loglike[i]
	#
	#
########### M-step ############
	# estimate mixing proportion
	prob<-sum(vi)/n
	#
	# maximization of each theta parameter
	if ( distribution == 1 ) {
		# calculation of theta_j, j=1,...,jmax ; theta_0=1-sum(theta)
#		print (c('1:discrete, jmax=', jmax))
		if (jmax==0){ theta<-0}
		else{
			for (ii in 1:jmax) {
				temp<-as.numeric(( (x==ii) & (y==ii) ));
				theta[ii]<-sum(temp*vi)/sum(vi) 
#				print( c(ii, sum(temp), sum(vi), sum(temp*vi) ) )
				}
#				print (c('2:discrete, jmax=', jmax))
			}
	}			
	else if (distribution==2){		
			# calculation of theta for poisson diagonal inflation
			theta<- sum(vi*x)/sum(vi) }
	else if (distribution==3){		
#	else {		
			# calculation of theta for geometric diagonal inflation
			theta<- sum(vi)/( sum(vi*x)+sum(vi) ) }
	# 	fit model on lambda3
	data2$v1<- 1-vi;
	data2$v1[ (data2$v1<0)&(data2$v1>-1.0e-10) ]<-0.0
#
	data2$y3<-s;
	m0<-glm( formula2, family=poisson, data=data2, weights=v1 , maxit=100)
	beta3<-m0$coef
	lambda3<-m0$fitted
	#
	#	fit model on lambda1 & lambda2
	data1$v1<- 1-c(vi,vi);
	data1$v1[ (data1$v1<0)&(data1$v1>-1.0e-10) ]<-0.0
#
	x1[(x1<0)&(x1>-1.0e-10)]<-0.0
	x2[(x2<0)&(x2>-1.0e-10)]<-0.0
	data1$y1y2<-c(x1,x2)
	m<-glm( formula1, family=poisson, data=data1, weights=v1 , maxit=100)
	p3<-length(m$coef)
	beta<-m$coef
#  ----	
#	creating names for parameters
	names(beta)<-newnamesbeta( beta )
#  ----	
	lambda<-fitted(m)
	lambda1<-lambda[1:n]
	lambda2<-lambda[(n+1):(2*n)]
	#
	#####   end of M step  ######
#
#	print all parameters including beta
	if (print.details) {
		printvec<- c( i,beta,beta3,100.0*prob, theta,  loglike[i], difllike );	
		names(printvec)<-c( 'iter', names(beta),paste('l3_',names(beta3),sep=''), 'Mix.p(%)', paste( 'theta', 1:length(theta),sep='' ), 'loglike', 'Rel.Dif.loglike')
		}
#
#  limited print out
	else {		
		printvec<- c( i, 100.0*prob, theta,  loglike[i], difllike );	
		names(printvec)<-c( 'iter', 'Mix.p(%)', paste( 'theta', 1:length(theta),sep='' ), 'loglike', 'Rel.Dif.loglike')
		}
#
	lengthprintvec<-length(printvec)
	print.default( printvec, digits=4 )
	}
#
#	calculation of BIC and AIC for bivpoisson model
	if ( (distribution==1)&&(jmax==0) ){noparams<- m$rank + m0$rank + 1}
	else                               {noparams<- m$rank + m0$rank + length( theta ) +1}
	AIC<- -2*loglike[i] + noparams * 2
	BIC<- -2*loglike[i] + noparams * log(2*n)
#	
#		
#	Calculation of BIC, AIC of Poisson saturated model
	x.mean<-x
	x.mean[x==0]<-1e-12
	y.mean<-y
	y.mean[y==0]<-1e-12
	AIC.sat <-  sum(log( dpois( x , x.mean ) ) + log( dpois( y , y.mean ) ))
	BIC.sat <-  -2 * AIC.sat + (2*n)* log(2*n)
	AIC.sat <-  -2 * AIC.sat + (2*n)* 2
#
#
	AICtotal<-c(AIC.sat, AIC); 
	BICtotal<-c(BIC.sat, BIC );
	names(AICtotal)<-c('Saturated', 'BivPois')
	names(BICtotal)<-c('Saturated', 'BivPois')
#  -------
#
#	spliting parameter vector
	betaparameters<-splitbeta( beta )
#
#	putting all betas in one vector
	allbeta<-c(betaparameters$beta1,betaparameters$beta2, beta3)
	names(allbeta)<-c( paste( '(l1):', names(betaparameters$beta1), sep='' ),paste('(l2):', names(betaparameters$beta2), sep='' ),paste('(l3):', names(beta3), sep='' ) )
#
# 	calculation of fitted values
#  ----------------------------
	fittedval1<-(1-prob)*(m$fitted[1:n]         + lambda3)
	fittedval2<-(1-prob)*(m$fitted[(n+1):(2*n)] + lambda3)
#  	
	meandiag<-0
	if ((distribution==1)&&(jmax>0)) { meandiag<-sum( theta[1:jmax]*1:jmax ) }
	else if (distribution==2) { meandiag<-theta }
	else if (distribution==3) { meandiag<- (1-theta)/theta }
#	
	fittedval1[x==y]<-prob*meandiag + fittedval1[x==y]
	fittedval2[x==y]<-prob*meandiag + fittedval2[x==y]
#
#
#	saving output
	result<-list(beta=allbeta, beta1=betaparameters$beta1, beta2=betaparameters$beta2, 
				beta3=beta3,  p=prob, theta=theta, diagonal.distribution=dilabel,
				lambda1=m$fitted[1:n], lambda2=m$fitted[(n+1):(2*n)], lambda3=lambda3, 
				fitted1=fittedval1, fitted2=fittedval2, loglikelihood=loglike[1:i], 
				parameters=noparams, AIC=AICtotal, 	BIC=BICtotal,iterations=i )
#
} # end of elseif
#
options(warn=0)
result
#
#
}

